require 'spec_helper'

describe Patriarch::Behaviours do


  describe '#add_behaviour' do

    let(:monster)      { Monster.create }
    let(:fallen_angel) { FallenAngel.create }
    let(:love_letter)  { LoveLetter.create }


    context 'on a bipartite' do

      before(:all) do
        Monster.add_behaviour       :like, :on => [:fallenAngel], :as => :love, :undo_as => :hate
        FallenAngel.add_behaviour   :like, :by => [:monster]    , :as => :love
      end

      context "call from a model including Patriarch::Behaviours" do

        it "defines instances methods to retrieve model instances having relationships with actor model" do
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_like")
        end

        it "defines instances methods to retrieve ids of model instances having relationships with actor model" do
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_like_ids")
        end      

        it "defines instances methods to retrieve model instances having relationships with target model" do
          progressive_like = (Verbs::Conjugator.conjugate :like, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_like}_me")
        end

        it "defines instances methods to retrieve ids of model instances having relationships with target model" do
          progressive_like = (Verbs::Conjugator.conjugate :like, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_like}_me_ids")
        end

        it "defines behaviour instance method on actor model" do
          monster.should          respond_to(:like)
        end
        
        it "defines undo behaviour instance method on actor model" do
          monster.should          respond_to(:undo_like)
        end

        it "does not define behaviour instance method on target model" do
          fallen_angel.should_not respond_to(:like)
        end

        it "does not define undo behaviour instance method on target model" do
          fallen_angel.should_not respond_to(:undo_like)
        end

        it "does not define alias for behaviour instance methods on target model" do
          progressive_love = (Verbs::Conjugator.conjugate :love, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_love}_me_ids")
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_love}_me")
        end

        it "defines alias for behaviour instance methods on actor model" do
          monster.should          respond_to(:love)
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_love")          
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_love_ids")
        end        
      end

      context 'implements methods that' do

        # Move tests in da right place ? Does it belongs to manager_service ?
        it 'lets transaction build without running' do
          monster.like(fallen_angel, :stop_execution => true).should be_an_instance_of(Patriarch::Transaction)
        end

        it 'lets undo transaction build without running' do
          monster.undo_like(fallen_angel, :stop_execution => true).should be_an_instance_of(Patriarch::Transaction)
        end      

       

        shared_examples 'instances have behaviour' do    
          it "remember target id in actor's memory" do 
            monster.fallen_angels_i_like_ids.should == [fallen_angel.id]
          end

          it "remember target instance in actor's memory" do 
            monster.fallen_angels_i_like.should == [fallen_angel]
          end
          
          it "remember actor id in target's memory" do 
            fallen_angel.monsters_liking_me_ids.should == [monster.id]        
          end

          it "remember actor instance in target's memory" do 
            fallen_angel.monsters_liking_me.should == [monster]
          end      
        end

        context 'when doing direct behaviour' do
          before(:all) { monster.like fallen_angel }  
          it_has_behavior 'instances have behaviour' 
          after(:all) { monster.undo_like fallen_angel }                
        end

        context 'when doing behaviour via alias method' do
          before(:all) { monster.love fallen_angel }       
          it_has_behavior 'instances have behaviour' 
          after(:all) { monster.undo_like fallen_angel }                
        end

        shared_examples 'instances have undo behaviour' do
          it "erase target id from actor's memory" do 
            monster.fallen_angels_i_like_ids.should be_empty
          end

          it "erase target instance from actor's memory" do 
            monster.fallen_angels_i_like.should be_empty
          end
          
          it "erase actor id from target's memory" do 
            fallen_angel.monsters_liking_me_ids.should be_empty    
          end

          it "erase actor instance from target's memory" do 
            fallen_angel.monsters_liking_me.should be_empty
          end             
        end

        context 'when calling undo command' do
          before(:all) do 
            monster.like fallen_angel  
            monster.undo_like fallen_angel 
          end              
          it_has_behavior 'instances have undo behaviour'
        end

        context 'when calling undo command via undo alias' do
          before(:all) do 
            monster.like fallen_angel  
            monster.hate fallen_angel 
          end          
          it_has_behavior 'instances have undo behaviour'
        end

      end 
    end # context bipartite

    context 'on a tripartite' do

      before(:all) do 
        # Monster is the actor model and FallenAngel the target model here
        Monster.add_behaviour       :praise, :on => [:fallenAngel], :via => [:loveLetter], :as => :flatter, :undo_as => :despise
        FallenAngel.add_behaviour   :praise, :by => [:monster], :via => [:loveLetter], :as => :flatter
        LoveLetter.add_behaviour    :praise, :medium_between => [:monster,:fallenAngel], :as => :flatter
      end

      context "call from a model including Patriarch::Behaviours" do
        it "defines instances methods to retrieve model instances having relationships with actor model" do
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_praise_via_#{"love_letter".pluralize}")
        end

        it "defines instances methods to retrieve ids of model instances having relationships with actor model" do
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_praise_via_#{"love_letter".pluralize}_ids")
        end      

        it "defines instances methods to retrieve instances of medium model instances used by actor model" do
          monster.should           respond_to("#{"love_letter".pluralize}_i_use_to_praise_#{"fallen_angel".pluralize}")
        end      

        it "defines instances methods to retrieve ids of medium model instances used by actor model" do
          monster.should           respond_to("#{"love_letter".pluralize}_i_use_to_praise_#{"fallen_angel".pluralize}_ids")
        end              

        it "defines instances methods to retrieve model instances having relationships with target model" do
          progressive_praise = (Verbs::Conjugator.conjugate :praise, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_praise}_me_via_#{"love_letter".pluralize}")
        end

        it "defines instances methods to retrieve ids of model instances having relationships with target model" do
          progressive_praise = (Verbs::Conjugator.conjugate :praise, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_praise}_me_via_#{"love_letter".pluralize}_ids")
        end

        it "defines instances methods to retrieve instances of medium model instances having relationships with target model" do  
          fallen_angel.should     respond_to("#{"love_letter".pluralize}_used_by_#{"monster".pluralize}_to_praise_me")       
        end

        it "defines instances methods to retrieve ids of medium model instances having relationships with target model" do  
          fallen_angel.should     respond_to("#{"love_letter".pluralize}_used_by_#{"monster".pluralize}_to_praise_me_ids")       
        end

        # TODO define_alias also ...
        it "defines instances methods to retrieve ids of model instances entities behaviouring via medium model" do
          pending('lazyness')
        end

        it "defines behaviour instance method on actor model" do
          monster.should          respond_to(:praise)
        end
        
        it "defines undo behaviour instance method on actor model" do
          monster.should          respond_to(:undo_praise)
        end

        it "does not define behaviour instance method on target model" do
          fallen_angel.should_not respond_to(:praise)
        end

        it "does not define undo behaviour instance method on target model" do
          fallen_angel.should_not respond_to(:undo_praise)
        end

        it "does not define behaviour instance method on medium model" do
          love_letter.should_not respond_to(:praise)
        end

        it "does not define undo behaviour instance method on medium model" do
          love_letter.should_not respond_to(:undo_praise)
        end  

        it "does not define alias for behaviour instance methods on target model" do
          progressive_flatter = (Verbs::Conjugator.conjugate :flatter, :aspect => :progressive).split(/ /).last
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_flatter}_me_via_#{"love_letter".pluralize}")
          fallen_angel.should     respond_to("#{"monster".pluralize}_#{progressive_flatter}_me_via_#{"love_letter".pluralize}_ids")
        end

        #TODO test alias correctly with receive rspec ...
        it "defines alias for behaviour instance methods on actor model" do
          monster.should          respond_to(:flatter)
          # from test just below
          monster.should          respond_to(:despise)          
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_flatter_via_#{"love_letter".pluralize}")    
          monster.should          respond_to("#{"fallen_angel".pluralize}_i_flatter_via_#{"love_letter".pluralize}_ids") 
        end 

        # SQLITE 3 IS SHITTING ALL OVER OUR TESTS, TODO FIXME 
        # :memory: option appears to have some limitations there
        it "defines alias on undo behaviour instance methods of actor model" do 
          #          monster.should          respond_to(:despise)
          pending "test is runned elsewhere, fix sqlite3 memory problem please"
        end
      end
      
      context 'implements methods that' do 
        it 'lets transaction build without running' do
          monster.praise(love_letter, fallen_angel, :stop_execution => true).should be_an_instance_of(Patriarch::Transaction)
        end

        it 'lets undo transaction build without running' do
          monster.praise(love_letter, fallen_angel, :stop_execution => true).should be_an_instance_of(Patriarch::Transaction)
        end

        # prepare examples to test both alias and normal methods ...
        shared_examples 'instances have behaviour working' do

          it "allows actor to remember id of the medium used to behave on target" do                 
            monster.love_letters_i_use_to_praise_fallen_angels_ids.should == [love_letter.id]
          end
          
          it "allows actor to remember instance of the medium used to behave on target" do
            monster.love_letters_i_use_to_praise_fallen_angels.should == [love_letter]
          end

          it "allows target to remember id of the medium used to behave on him" do
            fallen_angel.love_letters_used_by_monsters_to_praise_me_ids.should == [love_letter.id]
          end

          it "allows target to remember instance of the medium used to behave on him" do
            fallen_angel.love_letters_used_by_monsters_to_praise_me.should == [love_letter]
          end

          it "allows actor to remember target id" do 
            monster.fallen_angels_i_praise_via_love_letters_ids.should == [fallen_angel.id]
          end

          it "allows actor to remember target instance" do 
            monster.fallen_angels_i_praise_via_love_letters.should == [fallen_angel]
          end
          
          it "allows target to remember actor id" do 
            fallen_angel.monsters_praising_me_via_love_letters_ids.should == [monster.id]        
          end

          it "allows target to remember actor instance" do 
            fallen_angel.monsters_praising_me_via_love_letters.should == [monster]
          end         
        end

        context 'when doing direct behaviour' do
          before(:all) { monster.praise love_letter,fallen_angel }          
          it_has_behavior 'instances have behaviour working' 
          after(:all) { monster.undo_praise love_letter,fallen_angel }             
        end        

        context 'when doing behaviour via alias method' do 
          before(:all) { monster.flatter love_letter,fallen_angel }       
          it_has_behavior 'instances have behaviour working' 
          after(:all) { monster.undo_praise love_letter,fallen_angel }
        end         

        # prepare examples to test both undo alias and undo normal methods ...
        shared_examples 'instances have undo behaviour working' do

          it "erase medium id used to behave on target from actor's memory" do
            monster.love_letters_i_use_to_praise_fallen_angels_ids.should be_empty
          end

          it "erase medium instance used to behave on target from actor's memory" do
            monster.love_letters_i_use_to_praise_fallen_angels.should be_empty            
          end          

          it "erase medium id used to behave on target from target's memory" do
            fallen_angel.love_letters_used_by_monsters_to_praise_me_ids.should be_empty            
          end          

          it "erase medium instance used to behave on target from target's memory" do
            fallen_angel.love_letters_used_by_monsters_to_praise_me.should be_empty               
          end                   

          it "erase target id from actor's memory" do 
            monster.fallen_angels_i_praise_via_love_letters_ids.should be_empty
          end

          it "erase target instance from actor's memory" do 
            monster.fallen_angels_i_praise_via_love_letters.should be_empty
          end
          
          it "erase actor id from target's memory" do 
            fallen_angel.monsters_praising_me_via_love_letters_ids.should be_empty  
          end

          it "erase actor instance from target's memory" do 
            fallen_angel.monsters_praising_me_via_love_letters.should be_empty
          end         
        end

        context 'when directly undoing behaviour' do
          before(:all) do
            monster.praise love_letter, fallen_angel
            monster.undo_praise love_letter, fallen_angel
          end
          it_has_behavior 'instances have undo behaviour working' 
        end

        context 'when undoing behaviour via undo alias' do
          before(:all) do
            monster.praise love_letter,fallen_angel
            monster.despise love_letter,fallen_angel
          end
          it_has_behavior 'instances have undo behaviour working' 
        end

      end # implements methods that
    end # context tripartite

  end # desc '#add_behaviour'

  # FIXME learn rspec tocard ...
  context "multiple tricky calls" do 
    it "works" do
      class Monsterr < Monster; end
      class FallenAngell < FallenAngel; end
      Monsterr.add_behaviour       :praise, :on => [:fallenAngell], :via => [:fallenAngell]
      FallenAngell.add_behaviour   :praise, :by => [:monsterr],     :via => [:fallenAngell]
      FallenAngell.add_behaviour   :praise, :medium_between => [:monsterr,:fallenAngell]
    end
  end

  context 'when using behaviour on entities that do not allow it' do

    before(:all) do
      class Monsterr < Monster; end
      class FallenAngell < FallenAngel; end        
      class LoveLetterr < LoveLetter; end        
      Monsterr.add_behaviour       :like, :on => [:fallenAngell], :as => :love, :undo_as => :hate
      FallenAngell.add_behaviour   :like, :by => [:monsterr]    , :as => :love

      Monsterr.add_behaviour       :praise, :on => [:fallenAngell], :via => [:love_letter]
      FallenAngell.add_behaviour   :praise, :by => [:monsterr],     :via => [:love_letter]
      LoveLetterr.add_behaviour    :praise, :medium_between => [:monsterr,:fallenAngell]     
    end

    it "shall fail for bipartite" do
      expect{ Monsterr.create.like(LoveLetterr.create) }.to raise_error
    end

    it "shall fail for triparite with unauthorized medium class instances" do 
      expect{ Monsterr.create.praise(Monsterr.create,FallenAngell.create) }.to raise_error
    end

    it "shall fail for triparite with unauthorized target class instances" do 
      expect{ Monsterr.create.praise(LoveLetterr.create,Monsterr.create) }.to raise_error
    end    
  end

end # desc Patriarch::Behaviours
