require 'spec_helper'

describe 'entities including Patriarch::Behaviours #destroy' do

  before(:all) do
    Monster.add_behaviour       :lure, :by => [:fallenAngel]
    FallenAngel.add_behaviour   :lure, :on => [:monster]
    FallenAngel.add_behaviour   :lure, :by => [:fallenAngel]
    FallenAngel.add_behaviour   :lure, :on => [:fallenAngel]    
    Monster.add_behaviour       :like, :on => [:fallenAngel] 
    FallenAngel.add_behaviour   :like, :by => [:monster]

    # Tripartite behaviour ...
    FallenAngel.add_behaviour   :praise, :on => [:monster],     :via => [:loveLetter]
    Monster.add_behaviour       :praise, :by => [:fallenAngel], :via => [:loveLetter]
    LoveLetter.add_behaviour    :praise, :medium_between  =>  [:fallenAngel,:monster]
  end

  let(:monster)      { Monster.create }
  let(:beast)        { Monster.create }
  let(:love_letter)  { LoveLetter.create }

  it "performs redis clean callbacks when other callbacks are ok" do
    f = FallenAngel.create
    m = Monster.create
    f.lure m
    m.like f
    f.destroy
    f.destroyed?.should be_true
    f.monsters_i_lure.should be_empty
    f.monsters_liking_me.should be_empty
    m.fallen_angels_i_like.include?(f).should be_false
    m.fallen_angels_luring_me.include?(f).should be_false
  end

  context  "cascading style" do
    # KIKOOOOOOOO Test for Skelz0r
    before(:all) do
      FallenAngel.class_eval do
        before_destroy :destroy_fallen_angels_i_lure
        def destroy_fallen_angels_i_lure
          self.fallen_angels_i_lure.each do |lured_fallen_angel|
            lured_fallen_angel.destroy
          end  
        end
      end
    end 

    it "cleans safely on top of the 'transitive' behaviour chain" do
      f,fa,fal = [FallenAngel.create,FallenAngel.create,FallenAngel.create]
      f.lure fa; fa.lure fal
      
      f.destroy
      
      f.fallen_angels_i_lure.should be_empty
      fa.fallen_angels_i_lure.should be_empty
      
      fa.destroyed?.should be_false
      FallenAngel.where(:id => fa.id).should be_empty
      
      fal.destroyed?.should be_false
      FallenAngel.where(:id => fal.id).should be_empty    
      
      f.destroyed?.should be_true
      FallenAngel.where(:id => f.id).should be_empty    
    end

    it "cleans safely anywhere in the 'transitive' behaviour chain" do
      f,fa,fal = [FallenAngel.create,FallenAngel.create,FallenAngel.create]
      f.lure fa; fa.lure fal
      
      fa.destroy
      
      f.fallen_angels_i_lure.should be_empty
      fa.fallen_angels_i_lure.should be_empty
      
      fa.destroyed?.should be_true
      FallenAngel.where(:id => fa.id).should be_empty    

      fal.destroyed?.should be_false
      FallenAngel.where(:id => fa.id).should be_empty    

      f.destroyed?.should be_false
      Array(FallenAngel.where(:id => f.id)).should == [f]        
    end    
  end

  context ' when another callback fails' do
  
    # they do not trigger a rollback when they return false
    before(:all) do
      FallenAngel.class_eval do
        around_destroy  :callback_fail
        def callback_fail
          raise ActiveRecord::Rollback, "a callback failed" unless @redis_destroy_transac_object.nil?
        end
      end
    end
    
    # This works with ActiveRecord as expected
    it "prevents redis clean" do
      f = FallenAngel.create
      m = Monster.create
      f.lure m
      m.like f
      f.destroy
      f.monsters_i_lure.should == [m]
      f.monsters_liking_me.should == [m]
      m.fallen_angels_i_like.include?(f).should be_true
      m.fallen_angels_luring_me.include?(f).should be_true
    end
  end
  
end
